"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const precise_proofs_js_1 = require("precise-proofs-js");
const utils_general_1 = require("@energyweb/utils-general");
const __1 = require("..");
exports.getAllCertificateEvents = async (certId, configuration) => {
    const registry = configuration.blockchainProperties.certificateLogicInstance;
    const allEvents = await registry.getAllEvents({
        topics: [
            null,
            configuration.blockchainProperties.web3.utils.padLeft(configuration.blockchainProperties.web3.utils.fromDecimal(certId), 64, '0')
        ],
        fromBlock: 0,
        toBlock: 'latest'
    });
    const returnEvents = [];
    for (const fullEvent of allEvents) {
        // we have to remove some false positives due to ERC721 interface
        if (fullEvent.event === 'Transfer') {
            if (fullEvent.returnValues.tokenId === `${certId}`) {
                returnEvents.push(fullEvent);
            }
        }
        else {
            returnEvents.push(fullEvent);
        }
    }
    // we also have to search
    if (certId !== 0) {
        const transferEvents = await registry.getAllTransferSingleEvents({
            topics: [
                '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef',
                null,
                null,
                configuration.blockchainProperties.web3.utils.padLeft(configuration.blockchainProperties.web3.utils.fromDecimal(certId), 64, '0')
            ],
            fromBlock: 0,
            toBlock: 'latest'
        });
        for (const transferEvent of transferEvents) {
            returnEvents.push(transferEvent);
        }
    }
    return returnEvents;
};
class Entity extends utils_general_1.BlockchainDataModelEntity.Entity {
    constructor(id, configuration, isPrivate = false) {
        super(id, configuration);
        this.isPrivate = isPrivate;
        this.owners = {};
        this.initialized = false;
    }
    async sync() {
        if (this.id === null) {
            return this;
        }
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const certOnChain = await registry.getCertificate(Number(this.id));
        this.data = certOnChain.data;
        const issuer = this.configuration.blockchainProperties.issuerLogicInstance;
        const decodedData = await issuer.decodeData(this.data);
        const allIssuanceEvents = await registry.getAllIssuanceSingleEvents({
            filter: { _id: Number(this.id) }
        });
        const creationBlock = await this.configuration.blockchainProperties.web3.eth.getBlock(allIssuanceEvents[0].blockNumber);
        this.generationStartTime = Number(decodedData['0']);
        this.generationEndTime = Number(decodedData['1']);
        this.deviceId = decodedData['2'];
        this.issuer = certOnChain.issuer;
        this.creationTime = Number(creationBlock.timestamp);
        if (this.isPrivate) {
            this.ownershipCommitment = await this.getOffChainProperties();
            this.owners = {
                [this.ownershipCommitment.ownerAddress]: this.ownershipCommitment.volume
            };
        }
        else {
            this.owners = await this.calculateOwnership();
        }
        this.energy = Object.keys(this.owners).map(owner => this.owners[owner]).reduce((a, b) => a + b, 0);
        this.initialized = true;
        if (this.configuration.logger) {
            this.configuration.logger.verbose(`Certificate ${this.id} synced`);
        }
        return this;
    }
    async claim(amount) {
        const owner = this.configuration.blockchainProperties.activeUser.address;
        if (this.isPrivate) {
            const ownerBalance = this.ownedVolume();
            if (!ownerBalance || this.owners[owner] === null) {
                throw new Error(`transfer(): ${owner} does not own a share in certificate ${this.id}`);
            }
            if (amount && amount !== ownerBalance) {
                throw new Error(`transfer(): unable to claim amount ${amount} Wh. For private certificate you can only claim the full balance of ${ownerBalance} Wh`);
            }
            const issuer = this.configuration.blockchainProperties.issuerLogicInstance;
            const { salts } = await this.getOffChainDump();
            const calculatedOffChainStorageProperties = this.prepareEntityCreation(this.ownershipCommitment, __1.OwnershipCommitmentSchema, salts);
            const { logs } = await issuer.requestMigrateToPublic(Number(this.id), calculatedOffChainStorageProperties.leafs[1].hash);
            const requestId = this.configuration.blockchainProperties.web3.utils.hexToNumber(logs[0].topics[2]);
            const theProof = precise_proofs_js_1.PreciseProofs.createProof('ownerAddress', calculatedOffChainStorageProperties.leafs, false);
            const onChainProof = theProof.proofPath.map(p => ({
                left: !!p.left,
                hash: p.left || p.right
            }));
            await issuer.migrateToPublic(requestId, amount, calculatedOffChainStorageProperties.leafs[1].salt, onChainProof);
            this.isPrivate = false;
            amount = ownerBalance;
        }
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const { randomHex, hexToBytes } = this.configuration.blockchainProperties.web3.utils;
        // TO-DO: replace with proper claim data
        const claimData = hexToBytes(randomHex(32));
        return registry.safeTransferAndClaimFrom(owner, owner, parseInt(this.id, 10), amount, this.data, claimData, utils_general_1.Configuration.getAccount(this.configuration));
    }
    async transfer(to, amount) {
        const from = this.configuration.blockchainProperties.activeUser.address;
        if (this.isPrivate) {
            const senderBalance = this.ownedVolume();
            if (!senderBalance || this.owners[from] === null) {
                throw new Error(`transfer(): ${from} does not own a share in certificate ${this.id}`);
            }
            if (amount && amount !== senderBalance) {
                throw new Error(`transfer(): unable to send amount ${amount} Wh. For private certificate you can only send the full balance of ${senderBalance} Wh`);
            }
            const previousCommitment = this.propertiesDocumentHash;
            const ownershipCommitment = {
                ownerAddress: to,
                volume: this.ownedVolume()
            };
            const updatedOwnershipProperties = this.prepareEntityCreation(ownershipCommitment, __1.OwnershipCommitmentSchema);
            await this.syncOffChainStorage(ownershipCommitment, updatedOwnershipProperties);
            const issuer = this.configuration.blockchainProperties.issuerLogicInstance;
            const { logs } = await issuer.requestPrivateTransfer(Number(this.id), updatedOwnershipProperties.leafs[1].hash);
            const requestId = this.configuration.blockchainProperties.web3.utils.hexToNumber(logs[0].topics[2]);
            const theProof = precise_proofs_js_1.PreciseProofs.createProof('ownerAddress', updatedOwnershipProperties.leafs, false);
            const onChainProof = theProof.proofPath.map(p => ({
                left: !!p.left,
                hash: p.left || p.right
            }));
            await issuer.approvePrivateTransfer(requestId, onChainProof, previousCommitment, updatedOwnershipProperties.rootHash);
            this.propertiesDocumentHash = updatedOwnershipProperties.rootHash;
        }
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        return registry.safeTransferFrom(from, to, parseInt(this.id, 10), (amount !== null && amount !== void 0 ? amount : this.ownedVolume()), this.data, utils_general_1.Configuration.getAccount(this.configuration));
    }
    async revoke() {
        const issuer = this.configuration.blockchainProperties.issuerLogicInstance;
        return issuer.revokeCertificate(Number(this.id), utils_general_1.Configuration.getAccount(this.configuration));
    }
    async getAllCertificateEvents() {
        return exports.getAllCertificateEvents(parseInt(this.id, 10), this.configuration);
    }
    isOwned(byAddress) {
        const ownedVolume = this.ownedVolume(byAddress);
        return ownedVolume > 0;
    }
    ownedVolume(byAddress) {
        var _a;
        const owner = (byAddress !== null && byAddress !== void 0 ? byAddress : this.configuration.blockchainProperties.activeUser.address);
        return _a = this.owners[owner], (_a !== null && _a !== void 0 ? _a : 0);
    }
    async isClaimed(byAddress) {
        const claimedVolume = await this.claimedVolume(byAddress);
        return claimedVolume > 0;
    }
    async claimedVolume(byAddress) {
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const activeUserAddress = this.configuration.blockchainProperties.activeUser.address;
        const balance = await registry.claimedBalanceOf((byAddress !== null && byAddress !== void 0 ? byAddress : activeUserAddress), Number(this.id), utils_general_1.Configuration.getAccount(this.configuration));
        return Number(balance);
    }
    async calculateOwnership() {
        if (this.isPrivate) {
            throw new Error('Can only calculate ownership for public certificates');
        }
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        let transferSingleEvents = (await registry.getAllTransferSingleEvents({
            filter: { _id: this.id }
        }))
            .filter(e => e.returnValues._id === this.id)
            .map(e => e.returnValues);
        const transferBatchEvents = (await registry.getAllTransferBatchEvents()).map(e => e.returnValues);
        // Convert TransferBatch to TransferSingle event
        for (const event of transferBatchEvents.filter(e => e._ids.includes(this.id))) {
            for (let i = 0; i < event._ids.length; i++) {
                if (event._ids[i] === this.id) {
                    transferSingleEvents.push({
                        _id: event._ids[i],
                        _operator: event._operator,
                        _to: event._to,
                        _from: event._from,
                        _value: event._values[i]
                    });
                }
            }
        }
        const owners = {};
        for (const event of transferSingleEvents) {
            const { _from, _to, _value } = event;
            const valueTransferred = Number(_value);
            if (_from !== '0x0000000000000000000000000000000000000000') {
                owners[_from] = owners[_from] - valueTransferred;
            }
            if (owners[_to] === null || owners[_to] === undefined) {
                owners[_to] = 0;
            }
            owners[_to] += valueTransferred;
        }
        return owners;
    }
}
exports.Entity = Entity;
exports.createCertificate = async (to, value, fromTime, toTime, deviceId, configuration, isVolumePrivate = false) => {
    const newEntity = new Entity(null, configuration, isVolumePrivate);
    const getIdFromLogs = (logs) => configuration.blockchainProperties.web3.utils
        .hexToNumber(logs[0].topics[2])
        .toString();
    const issuer = configuration.blockchainProperties.issuerLogicInstance;
    const data = await issuer.encodeData(fromTime, toTime, deviceId);
    let tx;
    if (isVolumePrivate) {
        const ownershipCommitment = {
            ownerAddress: to,
            volume: value
        };
        const updatedOwnershipProperties = newEntity.prepareEntityCreation(ownershipCommitment, __1.OwnershipCommitmentSchema);
        await newEntity.syncOffChainStorage(ownershipCommitment, updatedOwnershipProperties);
        tx = await issuer.issuePrivate(to, updatedOwnershipProperties.rootHash, data, utils_general_1.Configuration.getAccount(configuration));
        newEntity.propertiesDocumentHash = updatedOwnershipProperties.rootHash;
    }
    else {
        tx = await issuer.issue(to, value, data, utils_general_1.Configuration.getAccount(configuration));
    }
    newEntity.id = getIdFromLogs(tx.logs);
    if (configuration.logger) {
        configuration.logger.info(`Certificate ${newEntity.id} created`);
    }
    return newEntity.sync();
};
async function claimCertificates(certificateIds, configuration) {
    const registry = configuration.blockchainProperties.certificateLogicInstance;
    const certificateIdsAsNumber = certificateIds.map(c => parseInt(c, 10));
    const certificatesPromises = certificateIds.map(certId => new Entity(certId, configuration).sync());
    const certificates = await Promise.all(certificatesPromises);
    const isOwnerPromise = certificates.map(cert => cert.isOwned());
    const owned = await Promise.all(isOwnerPromise);
    const ownsAllCertificates = owned.every(isOwned => isOwned === true);
    if (!ownsAllCertificates) {
        throw new Error(`You can only claim your own certificates`);
    }
    const valuesPromise = certificates.map(cert => cert.ownedVolume());
    const values = await Promise.all(valuesPromise);
    const { randomHex, hexToBytes } = configuration.blockchainProperties.web3.utils;
    // TO-DO: replace with proper claim data
    const claimData = certificates.map(cert => hexToBytes(randomHex(32)));
    const data = hexToBytes(randomHex(32));
    const { from } = utils_general_1.Configuration.getAccount(configuration);
    return registry.safeBatchTransferAndClaimFrom(from, from, certificateIdsAsNumber, values, data, claimData, utils_general_1.Configuration.getAccount(configuration));
}
exports.claimCertificates = claimCertificates;
async function transferCertificates(certificateIds, to, configuration) {
    const registry = configuration.blockchainProperties.certificateLogicInstance;
    const certificateIdsAsNumber = certificateIds.map(c => parseInt(c, 10));
    const certificatesPromises = certificateIds.map(certId => new Entity(certId, configuration).sync());
    const certificates = await Promise.all(certificatesPromises);
    const valuesPromise = certificates.map(cert => cert.ownedVolume());
    const values = await Promise.all(valuesPromise);
    const { randomHex, hexToBytes } = configuration.blockchainProperties.web3.utils;
    // TO-DO: replace with proper data
    const data = hexToBytes(randomHex(32));
    const { from } = utils_general_1.Configuration.getAccount(configuration);
    return registry.safeBatchTransferFrom(from, to, certificateIdsAsNumber, values, data, utils_general_1.Configuration.getAccount(configuration));
}
exports.transferCertificates = transferCertificates;
//# sourceMappingURL=Certificate.js.map